'use strict';

var config = require('../../config.json');
var doctrine = require('doctrine');

var WRONG_TYPES = {
    'object': 'Object',
    'array': 'Array',
    'number': 'Number',
    'string': 'String',
    'boolean': 'Boolean',
    'bool': 'Boolean',
    'any': '*'
};

/**
 * @param {String} name
 * @returns {String}
 */
function normalizeConstructorName(name) {
    if (WRONG_TYPES[name]) {
        return WRONG_TYPES[name];
    }

    return name;
}

/**
 * @param {Object} type
 * @returns {undefined|String|Array|Object}
 */
function normalizeTypes(type) {
    switch (type.type) {
        case 'AllLiteral':
            return void 0;

        case 'NullLiteral':
            return 'null';

        case 'UndefinedLiteral':
            return 'undefined';

        case 'NameExpression':
            return normalizeConstructorName(type.name);

        case 'UnionType':
            return type.elements.map(normalizeTypes);

        case 'TypeApplication':
            return {
                root: normalizeConstructorName(type.expression.name),
                children: type.applications.map(normalizeTypes)
            };

        case 'OptionalType':
            return {
                parameter: normalizeTypes(type.expression),
                optional: true
            };

        case 'RecordType':
            return {
                record: normalizeConstructorName('Object'),
                fields: type.fields.reduce(function (map, fieldType) {
                    map[fieldType.key] = normalizeTypes(fieldType);

                    return map;
                }, {})
            };

        case 'FieldType':
            return normalizeTypes(type.value);

        case 'NonNullableType':
        case 'NullableType':
            return normalizeTypes(type.expression);

        default:
            throw new Error('can\'t parse type ' + type.toString());
    }
}

/**
 * @param {String} comment - comment with jsDoc
 * @returns {{parameters: Object, returnStatement: Object, name: [String]}}
 */
module.exports = function parseDoc(comment) {
    /**
     * @type {{tags: Array<Object>, description: String}}
     */
    var commentAst = doctrine.parse(comment, config.doctrineConfig);

    if (!commentAst.tags.length) {
        return null;
    }

    var tags = commentAst.tags;
    var nameDescription = tags.find(function (tag) {
        return tag.title === 'name';
    });
    var paramsDescriptions = tags.filter(function (tag) {
        return tag.title === 'param';
    });
    var returnDescription = tags.find(function (tag) {
        return tag.title === 'return' || tag.title === 'returns';
    });

    var parameters = Object.create(null);
    var parameter = void 0;

    for (var index = 0, count = paramsDescriptions.length; index < count; index++) {
        parameter = paramsDescriptions[index];

        parameters[parameter.name] = normalizeTypes(parameter.type);
    }

    var returnStatement = void 0;

    if (returnDescription) {
        returnStatement = normalizeTypes(returnDescription.type);
    }

    var name = void 0;

    if (nameDescription) {
        name = nameDescription.name;
    }

    return {
        name: name,
        parameters: parameters,
        returnStatement: returnStatement
    };
};